# -*- coding: utf-8 -*-

# Le but de ce programme est de visualiser l'approximation d'une fonction 2π-périodique par sa série de Fourier 
# jusqu'à un certain rang.
# La principale modification par rapport au programme précédent consiste à implémenter une boucle supplémentaire, 
# afin d'afficher toutes les approximations jusqu'à un rang donné.

# On utilisera des séries de Fourier trigonométrique. La définition des coefficients de Fourier en est complexifiée,
# mais cela évite de passer par les complexes avant de revenir dans les réels.


import numpy as np
import matplotlib.pyplot as plt


# Tout d'abord, on définit quelques signaux : triangle, créneau, bizarre.
# Les fonctions ci-dessous définissent les signaux sur [-π,π], et sont étendues par périodicité.
# N'hésitez pas à implémenter les signaux de votre choix !

def signal_creneau (x) :
    if x > 0 :
        return 1
    return -1

def signal_triangle (x) :
    return np.pi/2 - abs(x)

def signal_bizarre (x) :
    return np.sin(x)*np.exp(np.cos(x+0.2))


# La fonction serie_Fourier calcule les coefficients de Fourier trigonométrique d'une fonction.
# Elle prend en entrée 3 variables :
#   * une fonction signal (définie sur [-π,π])
#   * un entier N (la fréquence la plus haute dont on calcule le coefficient de Fourier)
#   * un entier nombre_pas (le pas utilisé pour l'intégration numérique étant alors 2π/nombre_pas)
# Elle fournit 
#   * un réel (float), le coefficient 0 de la série de Fourier.
#   * une liste de réels de longueur N, les coefficients en cosinus de la série de Fourier.
#   * une liste de réels de longueur N, les coefficients en sinus de la série de Fourier.
# Le schéma d'intégration est une somme de Riemann classique. Celui-ci se comporte agréablement avec des fonctions périodiques.
# Il existe des algorithmes beaucoup plus efficaces pour cette tâche (transformée de Fourier rapide -- Fast Fourier Transform, FFT).

def serie_Fourier (signal, N, nombre_pas) :
    pas = 2*np.pi/nombre_pas # On pré-calcule le pas. Cela clarifie les calculs suivants.
    valeurs_signal = [signal(-np.pi + k*pas) for k in range(nombre_pas)] # On pré-calcule les valeurs de la fonction.
    C_0 = sum(valeurs_signal)/nombre_pas
    A = []
    for n in range (1, N+1) :
        valeurs_tordues = [valeurs_signal[k] * np.cos(n*(-np.pi+k*pas)) for k in range(nombre_pas)]
        A.append(2*sum(valeurs_tordues)/nombre_pas)
    B = []
    for n in range (1, N+1) :
        valeurs_tordues = [valeurs_signal[k] * np.sin(n*(-np.pi+k*pas)) for k in range(nombre_pas)]
        B.append(2*sum(valeurs_tordues)/nombre_pas)
    return C_0, A, B


# La fonction reconstruction_signal est la fonction reconstruite à partir des premiers coefficients de Fourier.
# Elle prend en entrée 4 variables :
#   * les coefficients de Fourier trigonométriques C_0, A, B
#   * un réel (float) x
# Elle fournit la valeur de la série de Fourier au rang N = len(A), évaluée en x.

def reconstruction_signal (C_0, A, B, x) :
    S = C_0
    N = len(A)
    for n in range(1, N+1) :
        S += A[n-1]*np.cos(n*x)
        S += B[n-1]*np.sin(n*x)
    return S


# La fonction graphique_decomposition_reconstruction affiche sur le même graphique le signal initial (en bleu)
# et les signaux reconstruits par série de Fourier (en dégradé).
# Elle prend en entrée 4 variables :
#   * une fonction signal (définie sur [-π,π])
#   * un entier N (la fréquence la plus haute dont on calcule le coefficient de Fourier)
#   * un entier nombre_pas (le pas utilisé pour l'intégration numérique étant alors 2π/nombre_pas, de même que le pas d'affichage)
# Elle affiche un graphique toutes les approximations jusqu'au rang N sur [-2π ,2π].

def graphique_decomposition_reconstruction_bis (signal, N, nombre_pas) :
    C_0, A, B = serie_Fourier (signal, N, nombre_pas)
    pas = 2*np.pi/nombre_pas
    signal_range = [-2*np.pi + k*pas for k in range(2*nombre_pas)]
    signal_image = [signal((k*pas+np.pi) % (2*np.pi) - np.pi) for k in range(2*nombre_pas)]
    plt.clf() # Nettoie la mémoire de matplotlib de ce qui pourrait traîner.
    plt.plot(signal_range, signal_image, c='blue')
    for n in range(N,-1,-1) : # On dessine les meilleures approximations en premier, pour des raisons esthétiques.
        signal_reconstruit_image = [reconstruction_signal(C_0, A[:n], B[:n], -2*np.pi + k*pas) for k in range(2*nombre_pas)]
        plt.plot(signal_range, signal_reconstruit_image, color=(1-n/(2*N),0,n/(2*N)))
            # La première approximation est en rouge, les suivantes tendent vers le bleu.
    plt.show() # Affiche le graphique.

# graphique_decomposition_reconstruction_bis (signal_creneau, 10, 100)
# graphique_decomposition_reconstruction_bis (signal_triangle, 10, 100)
graphique_decomposition_reconstruction_bis (signal_bizarre, 10, 100)

# Plus la fonction approchée est régulière, meilleure est l'approximation par des séries de Fourier.
# En effet, si la fonction approchée est très régulière, les coefficients de Fourier décroissent très vite,
# donc la perte causée par le fait que l'on coupe la série de Fourier au rang N est d'autant plus petite.

# On observe le phénomène de Gibbs pour la fonction créneau.
# La fonction "signal_bizarre" est tellement régulière que 7 coefficients (N=3) en fournissent déjà une très bonne approximation !