# -*- coding: utf-8 -*-

# Le but de ce programme est d'utiliser les séries de Fourier pour résoudre numériquement l'équation de la chaleur.
# On considère des conditions au bord 2π-périodiques.

# On utilisera des séries de Fourier trigonométrique. La définition des coefficients de Fourier en est complexifiée,
# mais cela évite de passer par les complexes avant de revenir dans les réels.


import numpy as np
import matplotlib.pyplot as plt


# Tout d'abord, on définit quelques conditions initiales : triangle, créneau, bizarre.
# Les fonctions ci-dessous définissent les conditions initiales sur [-π,π], et sont étendues par périodicité.
# N'hésitez pas à implémenter les conditions initiales de votre choix !

def signal_creneau (x) :
    if x > 0 :
        return 1
    return -1

def signal_triangle (x) :
    return np.pi/2 - abs(x)

def signal_bizarre (x) :
    return np.sin(x)*np.exp(np.cos(x+0.2))


# La fonction serie_Fourier calcule les coefficients de Fourier trigonométrique d'une fonction.
# Elle prend en entrée 2 variables :
#   * une fonction signal (définie sur [-π,π])
#   * un entier N (la fréquence la plus haute dont on calcule le coefficient de Fourier)
# Une troisième variable, nombre_pas, commande la précision de l'approximation numérique. Il est fixé par défaut à 100.
# Elle fournit 
#   * un réel (float), le coefficient 0 de la série de Fourier.
#   * une liste de réels de longueur N, les coefficients en cosinus de la série de Fourier.
#   * une liste de réels de longueur N, les coefficients en sinus de la série de Fourier.
# Le schéma d'intégration est une somme de Riemann classique. Celui-ci se comporte agréablement avec des fonctions périodiques.
# Il existe des algorithmes beaucoup plus efficaces pour cette tâche (transformée de Fourier rapide -- Fast Fourier Transform, FFT).

def serie_Fourier (signal, N, nombre_pas = 100) :
    pas = 2*np.pi/nombre_pas # On pré-calcule le pas. Cela clarifie les calculs suivants.
    valeurs_signal = [signal(-np.pi + k*pas) for k in range(nombre_pas)] # On pré-calcule les valeurs de la fonction.
    C_0 = sum(valeurs_signal)/nombre_pas
    A = []
    for n in range (1, N+1) :
        valeurs_tordues = [valeurs_signal[k] * np.cos(n*(-np.pi+k*pas)) for k in range(nombre_pas)]
        A.append(2*sum(valeurs_tordues)/nombre_pas)
    B = []
    for n in range (1, N+1) :
        valeurs_tordues = [valeurs_signal[k] * np.sin(n*(-np.pi+k*pas)) for k in range(nombre_pas)]
        B.append(2*sum(valeurs_tordues)/nombre_pas)
    return C_0, A, B


# La fonction reconstruction_signal est la fonction reconstruite à partir des premiers coefficients de Fourier.
# Elle prend en entrée 4 variables :
#   * les coefficients de Fourier trigonométriques C_0, A, B
#   * un réel (float) x
# Elle fournit la valeur de la série de Fourier au rang N = len(A), évaluée en x.

def reconstruction_signal (C_0, A, B, x) :
    S = C_0
    N = len(A)
    for n in range(1, N+1) :
        S += A[n-1]*np.cos(n*x)
        S += B[n-1]*np.sin(n*x)
    return S


# La fonction graphique_decomposition_reconstruction affiche sur le même graphique la condition initiale (en bleu)
# et la température jusqu'au temps T (en dégradé).
# Elle prend en entrée 3 variables :
#   * une condition initiale (définie sur [-π,π])
#   * un temps maximal T jusqu'auquel on résoud l'équation de la chaleur
#   * le nombre de pas temporels affichés n_T. On affiche donc la solution toutes les T/n_T unités de temps.
# Une quatrième variable, N, est l'ordre de la série de Fourier utilisée. Elle est fixé par défaut à 50.
# Une cinquième variable, nombre_pas, commande la précision de l'affichage. Elle est fixé par défaut à 100.
# Elle affiche un graphique de ces deux signaux sur [-2π ,2π].

def graphique_chaleur (signal, T, n_T, N = 50, nombre_pas = 100) :
    C_0, A, B = serie_Fourier (signal, N)
    pas = 2*np.pi/nombre_pas
    signal_range = [-2*np.pi + k*pas for k in range(2*nombre_pas)]
    plt.clf() # Nettoie la mémoire de matplotlib de ce qui pourrait traîner.
    for n in range(n_T+1) :
        t = n*T/n_T
        A_t = [A[k]*np.exp(-(k+1)**2*t) for k in range(N)]
        B_t = [B[k]*np.exp(-(k+1)**2*t) for k in range(N)]
        signal_reconstruit_image = [reconstruction_signal(C_0, A_t, B_t, -2*np.pi + k*pas) for k in range(2*nombre_pas)]
        plt.plot(signal_range, signal_reconstruit_image, color=(1-float(n)/n_T,0,float(n)/n_T))
            # La première approximation est en rouge, les suivantes tendent vers le bleu.
    plt.show() # Affiche le graphique.


# La condition initiale est la fonction créneau.
# On résoud l'équation de la chaleur jusqu'au temps 4.
# On affiche la température toutes les 4/40 = 1/10 unités de temps.

# graphique_chaleur (signal_creneau, 4, 40)


# A vous d'essayer avec d'autres conditions initiales !

# graphique_chaleur (signal_triangle, 4, 40)
# graphique_chaleur (signal_bizarre, 4, 40, N=4)
graphique_chaleur ((lambda x: 2*x), 4, 40)